// The front-end ('fe') code has the purpose of directing user traffic
// where we want it: that is, on a node that is alive and (possibly)
// not overloaded. We do this at two different levels, DNS and HTTP,
// with two slightly different targets: the former is focused on
// availability, while the latter attempts to evenly distribute
// resource usage.
//
// DNS is used to provide at least one address of an active server:
// the capability to return multiple results, and the high-ttl nature
// of the service mean that we can simply return all the active nodes
// on every request, maximizing the chances that at least one of them
// will be active over a longer period of time.
//
// HTTP requests are istantaneous, and we can't rely on the client
// doing retries, so we must point the user at a single, active node
// on every request. There are two different policies, depending on
// the type of the request:
//
// - SOURCE requests must always reach the current master node. Since
// redirects tend to confuse streaming sources, which might have very
// simple HTTP implementations, we simply proxy the stream to the
// master node.
//
// - listener requests must be routed to an available relay, taking
// utilization into account (be it in terms of bandwidth, cpu usage,
// or more). The fact that the DNS layer returns multiple addresses
// provides already a very rough form of load balancing, but for
// accurate bandwidth planning we can't just rely on clients'
// cooperation. So when a client requests a stream using its public
// URL we need to serve a redirect to the desired node, computed
// according to the load balancing policy. This is currently done by
// serving a M3U file pointing directly at the target node's icecast
// daemon (but this may lock clients to that specific target node on
// failure... client reconnection policies still need some
// investigation).
//
package fe
